"""
This file includes a set of helpers that make treating Swift as a filesystem a
bit easier.

See COPYING for license information.
"""
import datetime
import stat
import os
import urlparse
import time

from twisted.internet import defer, reactor, task
from twisted.internet.defer import succeed
from twisted.web.iweb import IBodyProducer, UNKNOWN_LENGTH
from twisted.internet.interfaces import IConsumer
from twisted.python import log

from zope import interface

from swftp.utils import OrderedDict
from swftp.utils import try_datetime_parse
from swftp.swift import NotFound, Conflict


def obj_to_path(path):
    " Convert an entire path to a (container, item) tuple "
    path = path.strip('/')
    path = urlparse.urljoin('/', path)
    path = path.strip('/')
    path_parts = path.split('/', 1)

    container = None
    if len(path_parts) > 0 and path_parts[0] != '':
        container = path_parts[0]
    item = None
    if len(path_parts) > 1:
        item = path_parts[1]

    return container, item


def cb_parse_account_headers(headers):
    return {
        'count': headers.get('x-account-container-count', 0),
        'size': headers.get('x-account-bytes-used', 0),
        'content_type': 'application/directory',
    }


def cb_parse_container_headers(headers):
    return {
        'count': headers.get('x-container-object-count', 0),
        'size': headers.get('x-container-bytes-used', 0),
        'content_type': 'application/directory',
    }


def cb_parse_object_headers(headers):
    return {
        'size': headers.get('content-length', 0),
        'last_modified': headers.get('last-modified', 0),
        'content_type': headers.get('content-type'),
    }


def swift_stat(last_modified=None, content_type="application/directory",
               count=1, bytes=0, size=0, **kwargs):
    size = int(size) or int(bytes)
    mtime = try_datetime_parse(last_modified)
    if not mtime:
        mtime = time.mktime(datetime.datetime.utcnow().timetuple())

    if content_type == "application/directory":
        mode = 755 | stat.S_IFDIR
    else:
        mode = 644 | stat.S_IFREG
    return os.stat_result((mode, 0, 0, count, 65535, 65535, size, mtime,
                           mtime, mtime))


class SwiftWriteFile(object):
    """ Adapts IBodyProducer and IConsumer """
    interface.implements(IBodyProducer, IConsumer)

    def __init__(self, length=None):
        self.length = length or UNKNOWN_LENGTH
        self.started = defer.Deferred()
        self.finished = defer.Deferred()
        self.consumer = None  # is set later
        self.producer = None  # is set later

    # IConsumer
    def registerProducer(self, producer, streaming):
        self.producer = producer
        assert streaming

    def unregisterProducer(self):
        self.finished.callback(None)

    def write(self, data):
        self.consumer.write(data)
        log.msg(metric='ingress_bytes', count=len(data))

    # IBodyProducer
    def startProducing(self, consumer):
        self.consumer = consumer

        def stopped(reason):
            return succeed(None)
        self.finished.addCallback(stopped)
        self.started.callback(self)
        return self.finished

    def pauseProducing(self):
        self.producer.pauseProducing()

    def resumeProducing(self):
        self.producer.resumeProducing()

    def stopProducing(self):
        self.producer.stopProducing()


class SwiftFileSystem(object):
    "Defines a common interface used to create Swift similar to a filesystem"
    def __init__(self, swiftconn):
        self.swiftconn = swiftconn

    def startFileUpload(self, fullpath):
        "returns IConsumer to write to object data to"
        container, path = obj_to_path(fullpath)
        consumer = SwiftWriteFile()
        d = self.swiftconn.put_object(container, path, body=consumer)
        return d, consumer

    def startFileDownload(self, fullpath, consumer, offset=0):
        "consumer: Protocol"
        container, path = obj_to_path(fullpath)
        headers = {}
        if offset > 0:
            headers['Range'] = 'bytes=%s-' % offset
        d = self.swiftconn.get_object(container, path, receiver=consumer,
                                      headers=headers)
        return d

    def touchFile(self, fullpath):
        container, path = obj_to_path(fullpath)
        return self.swiftconn.put_object(container, path, body=None)

    def checkFileExistance(self, fullpath):
        container, path = obj_to_path(fullpath)
        if container is None or path is None:
            raise NotImplementedError

        d = self.swiftconn.head_object(container, path)
        d.addCallback(cb_parse_object_headers)
        return d

    def removeFile(self, fullpath):
        container, path = obj_to_path(fullpath)
        if container is None or path is None:
            raise NotImplementedError
        d = self.swiftconn.delete_object(container, path)
        return d

    @defer.inlineCallbacks
    def renameFile(self, oldpath, newpath):
        container, path = obj_to_path(oldpath)
        newcontainer, newpath = obj_to_path(newpath)
        if not container or not newcontainer:
            raise NotImplementedError

        if not path and not newpath:
            # Attempt to 'rename' a container (metadata is lost)
            yield self.swiftconn.delete_container(container)
            yield self.swiftconn.put_container(newcontainer)
            defer.returnValue(None)
        else:
            # If the object doesn't actually exist, ABORT
            try:
                yield self.swiftconn.head_object(container, path)
            except NotFound:
                raise NotImplementedError

            # List out children of this path. If there are any, ABORT
            prefix = None
            if path:
                prefix = "%s/" % path
            _, children = yield self.swiftconn.get_container(
                container, prefix=prefix, limit=1)
            if len(children) > 0:
                raise NotImplementedError

            # This is an actual object with no children. Free to rename.
            yield self.swiftconn.put_object(
                newcontainer, newpath,
                headers={'X-Copy-From': '%s/%s' % (container, path)})
            yield self.swiftconn.delete_object(container, path)

    @defer.inlineCallbacks
    def getAttrs(self, fullpath):
        container, path = obj_to_path(fullpath)
        if path:
            try:
                headers = yield self.swiftconn.head_object(container, path)
                defer.returnValue(
                    cb_parse_object_headers(headers))
            except NotFound:
                prefix = None
                if path:
                    prefix = "%s/" % path
                _, children = yield self.swiftconn.get_container(
                    container, prefix=prefix, limit=1)
                if len(children) == 0:
                    raise NotFound(404, 'Not Found')
                defer.returnValue(
                    {'content_type': 'application/directory'})

        elif container:
            headers = yield self.swiftconn.head_container(container)
            defer.returnValue(cb_parse_container_headers(headers))
        else:
            headers = yield self.swiftconn.head_account()
            defer.returnValue(cb_parse_account_headers(headers))

    def makeDirectory(self, fullpath, attrs=None):
        container, path = obj_to_path(fullpath)
        if path:
            headers = {'Content-Type': 'application/directory'}
            d = self.swiftconn.put_object(container, path, headers=headers)
        else:
            d = self.swiftconn.put_container(container)
        return d

    @defer.inlineCallbacks
    def removeDirectory(self, fullpath):
        container, path = obj_to_path(fullpath)
        if path:
            yield self.swiftconn.delete_object(container, path)
        else:
            try:
                yield self.swiftconn.delete_container(container)
            except Conflict:
                # Wait 2 seconds and try to delete the container once more
                yield task.deferLater(
                    reactor, 2, self.swiftconn.delete_container, container)

    def get_full_listing(self, fullpath):
        """
            Return a full listing of objects, collapsed into a directory
            structure. Works for account, container and object prefix listings.


            @returns dict of {name: property} values
        """
        container, path = obj_to_path(fullpath)
        if container:
            return self.get_container_listing(container, path)
        else:
            return self.get_account_listing()

    def get_container_listing(self, container, path, marker=None,
                              all_files=None):
        if all_files is None:
            all_files = OrderedDict()
        prefix = None
        if path:
            prefix = "%s/" % path
        d = self.swiftconn.get_container(
            container, prefix=prefix, delimiter='/', marker=marker)

        def cb(results):
            r, files = results
            next_marker = None
            for f in files:
                if 'subdir' in f:
                    f['name'] = f['subdir']
                    f['content-type'] = 'application/directory'
                f['formatted_name'] = os.path.basename(
                    f['name'].encode("utf-8").rstrip('/'))
                all_files[f['formatted_name']] = f
                next_marker = f['name']
            if len(files) > 0:
                return self.get_container_listing(
                    container, path, marker=next_marker, all_files=all_files)
            return all_files
        d.addCallback(cb)
        return d

    def get_account_listing(self, marker=None, all_files=None):
        if all_files is None:
            all_files = OrderedDict()
        d = self.swiftconn.get_account(marker=marker)

        def cb(results):
            headers, files = results
            next_marker = None
            for f in files:
                f['content-type'] = 'application/directory'
                f['formatted_name'] = f['name'].encode("utf-8")
                all_files[f['formatted_name']] = f
                next_marker = f['name']
            if len(files) > 0:
                return self.get_account_listing(
                    marker=next_marker, all_files=all_files)
            return all_files
        d.addCallback(cb)
        return d



"""
Defines serviceMaker, which required for automatic twistd integration for
swftp-sftp

See COPYING for license information.
"""
from twisted.application.service import ServiceMaker

serviceMaker = ServiceMaker(
    'swftp-sftp',  # name
    'swftp.sftp.service',  # module
    'An SFTP Proxy Interface for Swift',  # description
    'swftp-sftp'  # tap name
)



"""
This file defines what is required for swftp-ftp to work with twistd.

See COPYING for license information.
"""
from swftp.utils import VERSION

from twisted.application import internet, service
from twisted.python import usage, log
from twisted.internet import reactor

import ConfigParser
import signal
import os
import sys


def run():
    options = Options()
    try:
        options.parseOptions(sys.argv[1:])
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)
    log.startLogging(sys.stdout)
    s = makeService(options)
    s.startService()
    reactor.run()


def get_config(config_path, overrides):
    defaults = {
        'auth_url': 'http://127.0.0.1:8080/auth/v1.0',
        'host': '0.0.0.0',
        'port': '5021',
        'num_persistent_connections': '4',
        'connection_timeout': '240',
        'verbose': 'false',
        'welcome_message': 'Welcome to SwFTP'
                           ' - an FTP interface for Openstack Swift',
        'log_statsd_host': '',
        'log_statsd_port': '8125',
        'log_statsd_sample_rate': '10.0',
        'log_statsd_metric_prefix': 'swftp.ftp',
    }
    c = ConfigParser.ConfigParser(defaults)
    c.add_section('ftp')
    if config_path:
        log.msg('Reading configuration from path: %s' % config_path)
        c.read(config_path)
    else:
        config_paths = [
            '/etc/swftp/swftp.conf',
            os.path.expanduser('~/.swftp.cfg')
        ]
        log.msg('Reading configuration from paths: %s' % config_paths)
        c.read(config_paths)
    for k, v in overrides.iteritems():
        if v:
            c.set('ftp', k, str(v))
    return c


class Options(usage.Options):
    "Defines Command-line options for the swftp-ftp service"
    optFlags = [
        ["verbose", "v", "Make the server more talkative"]
    ]
    optParameters = [
        ["config_file", "c", None, "Location of the swftp config file."],
        ["auth_url", "a", None,
            "Auth Url to use. Defaults to the config file value if it exists. "
            "[default: http://127.0.0.1:8080/auth/v1.0]"],
        ["port", "p", None, "Port to bind to."],
        ["host", "h", None, "IP to bind to."],
    ]


def makeService(options):
    """
    Makes a new swftp-ftp service. The only option is the config file
    location. The config file has the following options:
     - host
     - port
     - auth_url
     - num_persistent_connections
     - connection_timeout
     - welcome_message
    """
    from twisted.protocols.ftp import FTPFactory
    from twisted.web.client import HTTPConnectionPool
    from twisted.cred.portal import Portal

    from swftp.ftp.server import SwiftFTPRealm
    from swftp.auth import SwiftBasedAuthDB
    from swftp.utils import print_runtime_info

    print('Starting SwFTP-ftp %s' % VERSION)

    c = get_config(options['config_file'], options)
    ftp_service = service.MultiService()

    # Add statsd service
    if c.get('ftp', 'log_statsd_host'):
        try:
            from swftp.statsd import makeService as makeStatsdService
            makeStatsdService(
                c.get('ftp', 'log_statsd_host'),
                c.getint('ftp', 'log_statsd_port'),
                sample_rate=c.getfloat('ftp', 'log_statsd_sample_rate'),
                prefix=c.get('ftp', 'log_statsd_metric_prefix')
            ).setServiceParent(ftp_service)
        except ImportError:
            log.err('Missing Statsd Module. Requires "txstatsd"')

    pool = HTTPConnectionPool(reactor, persistent=True)
    pool.maxPersistentPerHost = c.getint('ftp', 'num_persistent_connections')
    pool.cachedConnectionTimeout = c.getint('ftp', 'connection_timeout')

    authdb = SwiftBasedAuthDB(auth_url=c.get('ftp', 'auth_url'),
                              verbose=c.getboolean('ftp', 'verbose'))

    ftpportal = Portal(SwiftFTPRealm())
    ftpportal.registerChecker(authdb)
    ftpfactory = FTPFactory(ftpportal)
    ftpfactory.welcomeMessage = c.get('ftp', 'welcome_message')
    ftpfactory.allowAnonymous = False

    signal.signal(signal.SIGUSR1, print_runtime_info)
    signal.signal(signal.SIGUSR2, print_runtime_info)

    internet.TCPServer(
        c.getint('ftp', 'port'),
        ftpfactory,
        interface=c.get('ftp', 'host')).setServiceParent(ftp_service)
    return ftp_service



"""
Deals with file upload and download. This streams to and from the OpenStack
Swift server.

See COPYING for license information.
"""
from zope import interface

from twisted.internet import defer, task, reactor
from twisted.conch.ssh.filetransfer import (
    FXF_CREAT, FXF_TRUNC, SFTPError, FX_NO_SUCH_FILE, FX_FAILURE,
    FX_CONNECTION_LOST, FXF_WRITE)
from twisted.conch.interfaces import ISFTPFile
from twisted.internet.protocol import Protocol
from twisted.internet.interfaces import IPushProducer
from twisted.internet.error import ConnectionLost
from twisted.python import log

from swftp.swift import NotFound


def cb_log_egress_bytes(result):
    if result:
        log.msg(metric='egress_bytes', count=len(result))
    return result


class SwiftFileReceiver(Protocol):
    "Streams data from Swift user to SFTP session"
    download_buffer_limit = 1024 * 1024
    upload_buffer_limit = 1024 * 1024

    def __init__(self, size, session):
        self.size = size
        self.session = session
        self.finished = defer.Deferred()
        self.done = False
        self.consume_paused = False

        self._offset = 0
        self._recv_buffer = ""
        self._recv_listeners = []
        self.transport = None

    def dataReceived(self, bytes):
        """
            Data has been received from Swift. Pauses Swift if the
            download_buffer_limit has been reached.
        """
        self._recv_buffer += bytes
        self._readloop()
        if len(self._recv_buffer) > self.download_buffer_limit:
            self.consume_paused = True
            self.transport.pauseProducing()

    def _checksessionbuffertimer(self):
        """
            Checks session buffer to see if we need to resume.
            Reschedules itself if the buffer is still not small enough
        """
        if not self.consume_paused:
            return
        if len(self.session.buf) <= self.upload_buffer_limit:
            self.consume_paused = False
            self.transport.resumeProducing()
        else:
            reactor.callLater(0, self._checksessionbuffertimer)

    def _checksessionbuffer(self):
        "Checks buffer size to see if we need to pause"
        if not self.transport:
            return
        if self.consume_paused:
            return
        if len(self.session.buf) > self.upload_buffer_limit:
            self.consume_paused = True
            self.transport.pauseProducing()
            reactor.callLater(0, self._checksessionbuffertimer)

    def _readloop(self):
        """
            The loop that checks to see if there is enough data to give back to
            the SFTP client.
        """
        self._checksessionbuffer()
        for callback in self._recv_listeners:
            d, offset, length = callback
            if len(self._recv_buffer) >= length:
                data = self._recv_buffer[:length]
                d.callback(data)
                self._recv_listeners.remove(callback)
                self._offset += len(data)
                self._recv_buffer = self._recv_buffer[length:]

                if self.consume_paused and \
                        len(self._recv_buffer) <= self.download_buffer_limit:
                    self.consume_paused = False
                    self.transport.resumeProducing()
            else:
                break

    def read(self, offset, length):
        """
            Register the fact that this session wants a slice of data
            described by the given offset/length. Returns a deferred that fires
            with the data once it is available.
        """
        if offset + length > self.size:
            length = self.size - offset

        def cb(result):
            if result is None:
                raise EOFError("EOF")
            return result

        # It looks like the SFTP client is asking for too much.
        if self.done and len(self._recv_buffer) == 0:
            raise EOFError("EOF")

        d = defer.Deferred()
        d.addCallback(cb)
        self._recv_listeners.append((d, offset, length))
        self._readloop()
        return d

    def connectionLost(self, reason):
        """
            For some reason, the HTTP connection has been lost. We can either
            be done reading from Swift or something back could have happened.
        """
        from twisted.web._newclient import ResponseDone
        from twisted.web.http import PotentialDataLoss

        self.done = True

        if reason.check(ResponseDone) or reason.check(PotentialDataLoss):
            self._readloop()
            for callback in self._recv_listeners:
                d, offset, length = callback
                d.errback(reason)
            self._recv_listeners = []
            self.finished.callback(None)
        else:
            for callback in self._recv_listeners:
                d, offset, length = callback
                d.errback(reason)
            self._recv_listeners = []
            self.finished.errback(reason)


class SwiftFileSender(object):
    "Streams data from SFTP user to Swift"
    interface.implements(IPushProducer)

    def __init__(self, swiftfilesystem, fullpath):
        self.swiftfilesystem = swiftfilesystem
        self.fullpath = fullpath

        self.write_finished = None  # Deferred that fires when finished writing
        self._task = None           # Task loop
        self._done_sending = False  # Set to True when the user closes the file
        self._writeBuffer = []

        self.started = False

    def pauseProducing(self):
        self._task.pause()

    def resumeProducing(self):
        self._task.resume()

    def stopProducing(self):
        if self._task:
            try:
                self._task.stop()
            except task.TaskStopped:
                pass

        for buf in self._writeBuffer:
            d, data = buf
            d.errback(SFTPError(FX_CONNECTION_LOST, 'Connection Lost'))
            self._writeBuffer.remove(buf)

    def _writeFlusher(self, writer):
        while True:
            if self._done_sending and len(self._writeBuffer) == 0:
                writer.unregisterProducer()
                break

            if len(self._writeBuffer) == 0:
                yield
                continue

            try:
                d, data = self._writeBuffer.pop(0)
                writer.write(data)
                d.callback(len(data))
                yield
            except IndexError:
                pass
            finally:
                yield

    def cb_start_task(self, writer):
        self._task = task.cooperate(self._writeFlusher(writer))

    def close(self):
        self._done_sending = True
        return self.write_finished

    def write(self, data):
        if not self.started:
            # If we haven't started uploading to Swift, start up that process
            self.write_finished, writer = \
                self.swiftfilesystem.startFileUpload(self.fullpath)
            writer.registerProducer(self, streaming=True)
            writer.started.addCallback(self.cb_start_task)
            self.started = True
        d = defer.Deferred()
        self._writeBuffer.append((d, data))
        return d


class SwiftFile(object):
    "Acts as an open file for the SFTP Server instance"
    interface.implements(ISFTPFile)

    def __init__(self, server, fullpath, flags=None, attrs=None):
        self.server = server
        self.swiftfilesystem = server.swiftfilesystem
        self.fullpath = fullpath
        self.flags = flags
        self.attrs = attrs
        self.r = None
        self.w = None
        self.props = None

    def checkExistance(self):
        """
            Checks whether or not the file exists. If the file flags specify,
            it will create the file and return a deffered with that has been
            completed.
        """
        d = self.swiftfilesystem.checkFileExistance(self.fullpath)

        def cb(props):
            self.props = props

        def errback(failure):
            failure.trap(NotFound)
            if self.flags & FXF_CREAT == FXF_CREAT:
                return self.swiftfilesystem.touchFile(self.fullpath)
            if self.flags & FXF_TRUNC == FXF_TRUNC:
                return self.swiftfilesystem.touchFile(self.fullpath)
            else:
                raise SFTPError(FX_NO_SUCH_FILE, 'File Not Found')

        if self.flags & FXF_WRITE == FXF_WRITE:
            self.w = SwiftFileSender(self.swiftfilesystem, self.fullpath)
        d.addCallback(cb)
        d.addErrback(errback)
        return d

    # New Writer Methods
    def close(self):
        " Returns a deferred that fires when the connection is closed "
        if self.w:
            d = defer.maybeDeferred(self.w.close)
            d.addErrback(self._errClose)
            return d

    def _errClose(self, failure):
        failure.trap(ConnectionLost, NotFound)
        if failure.check(ConnectionLost):
            raise SFTPError(FX_CONNECTION_LOST, "Connection Lost")
        elif failure.check(NotFound):
            raise SFTPError(FX_FAILURE, "Container Doesn't Exist")

    def writeChunk(self, offset, data):
        d = self.w.write(data)

        def errback(failure):
            raise SFTPError(FX_FAILURE, 'Upload Failure')
        d.addErrback(errback)
        return d

    # Reading Methods
    def readChunk(self, offset, length):
        if not self.r:
            self.r = SwiftFileReceiver(
                int(self.props['size']), session=self.session)
            self.swiftfilesystem.startFileDownload(
                self.fullpath, self.r, offset=offset)
        d = self.r.read(offset, length)
        d.addCallback(cb_log_egress_bytes)
        return d

    def getAttrs(self):
        return self.server.getAttrs(self.fullpath)

    def setAttrs(self, attrs):
        raise NotImplementedError



from collections import defaultdict

from twisted.internet import reactor, tcp
from twisted.python import log
from txstatsd.client import TwistedStatsDClient, StatsDClientProtocol  # NOQA
from txstatsd.metrics.metrics import Metrics
from txstatsd.process import PROCESS_STATS, NET_STATS, COUNTER_STATS
from txstatsd.report import ReportingService


def makeService(host='127.0.0.1', port=8125, sample_rate=1.0, prefix=''):
    client = TwistedStatsDClient(host, port)
    metrics = Metrics(connection=client, namespace=prefix)
    reporting = ReportingService()

    for report in PROCESS_STATS:
        reporting.schedule(report, sample_rate, metrics.gauge)

    for report in NET_STATS:
        reporting.schedule(report, sample_rate, metrics.gauge)

    for report in COUNTER_STATS:
        reporting.schedule(report, sample_rate, metrics.gauge)

    # Attach statsd log observer
    metric_collector = StatsdMetricCollector(metrics)
    reporting.schedule(metric_collector.report_metrics, sample_rate, None)
    log.addObserver(metric_collector.emit)

    protocol = StatsDClientProtocol(client)
    reactor.listenUDP(0, protocol)
    return reporting


class StatsdMetricCollector(object):
    def __init__(self, metric):
        self.metric = metric
        self.metrics = defaultdict(int)

    def emit(self, eventDict):
        if 'metric' in eventDict:
            self.metrics[eventDict['metric']] += eventDict.get('count', 1)

    def reset_metrics(self):
        self.metrics = defaultdict(int)

    def report_metrics(self):
        # Report collected metrics
        results = self.metrics
        self.reset_metrics()
        for name, value in results.items():
            self.metric.increment(name, value)

        # Generate/send Aux stats
        num_clients = len(
            [r for r in reactor.getReaders() if isinstance(r, tcp.Server)])
        self.metric.gauge('clients', num_clients)



"""
swift includes a basic swift client for twisted.

    client = SwiftConnection(auth_url, username, api_key, pool=pool)
    d = client.put_object('container', 'path/to/obj')

See COPYING for license information.
"""
from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.web.client import Agent, WebClientContextFactory
from twisted.internet.protocol import Protocol
from twisted.web.http_headers import Headers
from twisted.web import error
from twisted.web.iweb import IBodyProducer
from twisted.web._newclient import ResponseDone
from twisted.web.http import PotentialDataLoss
from twisted.python import log

from zope.interface import implements

import json
from urllib import quote as _quote


class StringProducer(object):
    implements(IBodyProducer)

    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class RequestError(error.Error):
    pass


class NotFound(RequestError):
    pass


class UnAuthenticated(RequestError):
    pass


class UnAuthorized(RequestError):
    pass


class Conflict(RequestError):
    pass


class ResponseReceiver(Protocol):
    """
    Assembles HTTP response from return stream.
    """

    def __init__(self, finished):
        self.recv_chunks = []
        self.finished = finished

    def dataReceived(self, bytes, final=False):
        self.recv_chunks.append(bytes)

    def connectionLost(self, reason):

        if reason.check(ResponseDone) or reason.check(PotentialDataLoss):
            self.finished.callback(''.join(self.recv_chunks))
        else:
            self.finished.errback(reason)


class ResponseIgnorer(Protocol):
    def __init__(self, finished):
        self.finished = finished

    def makeConnection(self, transport):
        transport.stopProducing()
        self.finished.callback(None)

    def dataReceived(self, bytes):
        pass

    def connectionLost(self, reason):
        pass


def cb_recv_resp(response, load_body=False, receiver=None):
    d_resp_recvd = Deferred()
    if response.code == 204:
        response.deliverBody(ResponseIgnorer(d_resp_recvd))
    elif load_body:
        response.deliverBody(ResponseReceiver(d_resp_recvd))
    else:
        if receiver:
            response.deliverBody(receiver)
            return
        else:
            response.deliverBody(ResponseIgnorer(d_resp_recvd))
    return d_resp_recvd.addCallback(cb_process_resp, response)


def cb_process_resp(body, response):
    # Emulate HTTPClientFactory and raise t.w.e.Error if we have errors.
    if response.code == 404:
        raise NotFound(response.code, body)
    if response.code == 401:
        raise UnAuthenticated(response.code, body)
    if response.code == 403:
        raise UnAuthorized(response.code, body)
    if response.code == 409:
        raise Conflict(response.code, body)
    elif response.code > 299 and response.code < 400:
        raise error.PageRedirect(response.code, body)
    elif response.code > 399:
        raise RequestError(response.code, body)
    headers = {}
    for k, v in response.headers.getAllRawHeaders():
        headers[k.lower()] = v.pop()
    response.headers = headers
    return response, body


def format_head_response(result):
    resp, body = result
    return resp.headers


def cb_json_decode(result):
    resp, body = result
    return resp, json.loads(body)


class SwiftConnection:
    """
        A basic connection class to interface with OpenStack Swift.
    """
    user_agent = 'Twisted Swift'

    def __init__(self, auth_url, username, api_key, pool=None, verbose=False):
        self.auth_url = auth_url
        self.username = username
        self.api_key = api_key
        self.storage_url = None
        self.auth_token = None
        self.contextFactory = WebClientContextFactory()
        self.contextFactory.noisy = False
        self.pool = pool
        self.verbose = verbose
        # self.agent = Agent(reactor, self.contextFactory, pool=pool)

    def make_request(self, method, path, params=None, headers=None, body=None,
                     body_reader=None):
        h = {
            'User-Agent': [self.user_agent],
        }
        if headers:
            for k, v in headers.iteritems():
                if not isinstance(v, list):
                    h[k] = [v]
                else:
                    h[k] = v
        url = "/".join((self.storage_url, path))
        if params:
            param_lst = []
            for k, v in params.iteritems():
                param_lst.append("%s=%s" % (k, v))
            url = "%s?%s" % (url, "&".join(param_lst))
        agent = Agent(reactor, self.contextFactory, pool=self.pool)

        def doRequest(ignored):
            h['X-Auth-Token'] = [self.auth_token]
            if self.verbose:
                log.msg('Request: %s %s, headers: %s' % (method, url, h))
            return agent.request(method, url, Headers(h), body)

        d = doRequest(None)

        def retryAuth(response):
            if response.code in [401, 403]:
                d_resp_recvd = Deferred()
                response.deliverBody(ResponseIgnorer(d_resp_recvd))
                d_resp_recvd.addCallback(self.cb_retry_auth)
                d_resp_recvd.addCallback(doRequest)
                return d_resp_recvd
            return response
        d.addCallback(retryAuth)

        return d

    def cb_retry_auth(self, ignored):
        return self.authenticate()

    def after_authenticate(self, result):
        response, body = result
        self.storage_url = response.headers['x-storage-url']
        self.auth_token = response.headers['x-auth-token']
        return result

    def authenticate(self):
        headers = {
            'User-Agent': [self.user_agent],
            'X-Auth-User': [self.username],
            'X-Auth-Key': [self.api_key],
        }
        agent = Agent(reactor, self.contextFactory, pool=self.pool)
        d = agent.request('GET', self.auth_url, Headers(headers))
        d.addCallback(cb_recv_resp, load_body=True)
        d.addCallback(self.after_authenticate)
        return d

    def head_account(self):
        d = self.make_request('HEAD', '')
        d.addCallback(cb_recv_resp)
        d.addCallback(format_head_response)
        return d

    def get_account(self, marker=None):
        params = {'format': 'json'}
        if marker:
            params['marker'] = quote(marker)
        d = self.make_request('GET', '', params=params)
        d.addCallback(cb_recv_resp, load_body=True)
        d.addCallback(cb_json_decode)
        return d

    def head_container(self, container):
        d = self.make_request('HEAD', quote(container))
        d.addCallback(cb_recv_resp)
        d.addCallback(format_head_response)
        return d

    def get_container(self, container, marker=None, prefix=None, path=None,
                      delimiter=None, limit=None):
        params = {'format': 'json'}
        if marker:
            params['marker'] = quote(marker)
        if path:
            params['path'] = quote(path)
        if prefix:
            params['prefix'] = quote(prefix)
        if delimiter:
            params['delimiter'] = quote(delimiter)
        if limit:
            params['limit'] = str(limit)
        d = self.make_request('GET', quote(container), params=params)
        d.addCallback(cb_recv_resp, load_body=True)
        d.addCallback(cb_json_decode)
        return d

    def put_container(self, container, headers=None):
        d = self.make_request('PUT', quote(container),
                              headers=headers)
        d.addCallback(cb_recv_resp)
        return d

    def delete_container(self, container):
        d = self.make_request('DELETE', quote(container))
        d.addCallback(cb_recv_resp)
        return d

    def head_object(self, container, path):
        _path = "/".join((quote(container), quote(path)))
        d = self.make_request('HEAD', _path)
        d.addCallback(cb_recv_resp)
        d.addCallback(format_head_response)
        return d

    def get_object(self, container, path, headers=None, receiver=None):
        _path = "/".join((quote(container), quote(path)))
        d = self.make_request('GET', _path, headers=headers)
        d.addCallback(cb_recv_resp, receiver=receiver)
        return d

    def put_object(self, container, path, headers=None, body=None):
        if not headers:
            headers = {}
        if not body:
            headers['Content-Length'] = '0'
        _path = "/".join((quote(container), quote(path)))
        d = self.make_request('PUT', _path, headers=headers, body=body)
        d.addCallback(cb_recv_resp, load_body=True)
        return d

    def delete_object(self, container, path):
        _path = "/".join((quote(container), quote(path)))
        d = self.make_request('DELETE', _path)
        d.addCallback(cb_recv_resp)
        return d


def quote(value, safe='/'):
    """
    Patched version of urllib.quote that encodes utf8 strings before quoting
    """
    value = encode_utf8(value)
    if isinstance(value, str):
        return _quote(value, safe)
    else:
        return value


def encode_utf8(value):
    if isinstance(value, unicode):
        value = value.encode('utf8')
    return value



"""
This file contains the primary server code for the FTP server.

See COPYING for license information.
"""
from zope.interface import implements
from twisted.cred import portal
from twisted.protocols.ftp import IFTPShell, IReadFile, IWriteFile, \
    FileNotFoundError, CmdNotImplementedForArgError, IsNotADirectoryError, \
    IsADirectoryError
from twisted.internet import defer
from twisted.internet.protocol import Protocol
from twisted.python import log
import stat

from swftp.swiftfilesystem import SwiftFileSystem, swift_stat, obj_to_path
from swftp.swift import NotFound, Conflict


class SwiftFTPRealm:
    implements(portal.IRealm)

    def getHomeDirectory(self):
        return '/'

    def requestAvatar(self, avatarId, mind, *interfaces):
        shell = SwiftFTPShell(avatarId)
        return interfaces[0], shell, shell.logout


def stat_format(keys, props):
    st = swift_stat(**props)
    l = []
    for key in keys:
        if key == 'size':
            val = st.st_size
        elif key == 'directory':
            val = st.st_mode & stat.S_IFDIR == stat.S_IFDIR
        elif key == 'permissions':
            val = st.st_mode
        elif key == 'hardlinks':
            val = 0
        elif key == 'modified':
            val = int(st.st_mtime)
        elif key in 'owner':
            val = 'nobody'
        elif key in 'group':
            val = 'nobody'
        else:  # Unknown Value
            val = ''
        l.append(val)
    return l


class SwiftFTPShell:
    """ Implements all the methods needed to treat Swift as an FTP Shell """
    implements(IFTPShell)

    def __init__(self, swiftconn):
        self.swiftconn = swiftconn
        self.swiftfilesystem = SwiftFileSystem(self.swiftconn)
        self.log_command('login')

    def log_command(self, command, *args):
        arg_list = ', '.join(str(arg) for arg in args)
        log.msg("COMMAND: %s(%s)" % (command, arg_list),
                system="SwFTP-FTP, (%s)" % self.swiftconn.username,
                metric='command.%s' % command)

    def logout(self):
        self.log_command('logout')

    def _fullpath(self, path_parts):
        return '/'.join(path_parts)

    def makeDirectory(self, path):
        self.log_command('makeDirectory', path)
        fullpath = self._fullpath(path)
        return self.swiftfilesystem.makeDirectory(fullpath)

    def removeDirectory(self, path):
        self.log_command('removeDirectory', path)
        fullpath = self._fullpath(path)

        def not_found_eb(failure):
            failure.trap(NotFound)

        def conflict_eb(failure):
            failure.trap(Conflict)
            raise CmdNotImplementedForArgError(
                'Cannot delete non-empty directories.')

        d = self.swiftfilesystem.removeDirectory(fullpath)
        d.addErrback(not_found_eb)
        d.addErrback(conflict_eb)
        return d

    def removeFile(self, path):
        self.log_command('removeFile', path)
        fullpath = self._fullpath(path)

        def errback(failure):
            failure.trap(NotFound)
        d = self.swiftfilesystem.removeFile(fullpath)
        d.addErrback(errback)
        return d

    def rename(self, fromPath, toPath):
        self.log_command('rename', fromPath, toPath)
        oldpath = self._fullpath(fromPath)
        newpath = self._fullpath(toPath)

        d = self.swiftfilesystem.renameFile(oldpath, newpath)

        def errback(failure):
            failure.trap(NotFound, Conflict, NotImplementedError)
            if failure.check(NotFound):
                return defer.fail(FileNotFoundError(oldpath))
            else:
                return defer.fail(CmdNotImplementedForArgError(oldpath))
        d.addErrback(errback)
        return d

    def access(self, path):
        self.log_command('access', path)
        fullpath = self._fullpath(path)

        d = self.swiftfilesystem.getAttrs(fullpath)

        def cb(result):
            if result['content_type'] == 'application/directory':
                return defer.succeed(lambda: None)
            return defer.fail(IsNotADirectoryError(fullpath))
        d.addCallback(cb)

        def err(failure):
            failure.trap(NotFound)
            # Containers need to actually exist before uploading anything
            # inside of them. Therefore require containers to actually exist.
            # All other paths don't have to.
            if len(path) != 1:
                return defer.succeed(lambda: None)
            else:
                return defer.fail(IsNotADirectoryError(fullpath))

        d.addErrback(err)
        return d

    def stat(self, path, keys=()):
        self.log_command('stat', path, keys)
        fullpath = self._fullpath(path)

        def cb(result):
            return stat_format(keys, result)

        def err(failure):
            failure.trap(NotFound)
            return defer.fail(FileNotFoundError(fullpath))

        d = self.swiftfilesystem.getAttrs(fullpath)
        d.addCallback(cb)
        d.addErrback(err)
        return d

    def list(self, path=None, keys=()):
        self.log_command('list', path)
        fullpath = self._fullpath(path)

        def cb(results):
            l = []
            for key, value in results.iteritems():
                l.append([key, stat_format(keys, value)])
            return l

        def err(failure):
            failure.trap(NotFound)
            return defer.fail(FileNotFoundError(fullpath))

        d = self.swiftfilesystem.get_full_listing(fullpath)
        d.addCallback(cb)
        d.addErrback(err)
        return d

    def openForReading(self, path):
        self.log_command('openForReading', path)
        fullpath = self._fullpath(path)

        def cb(results):
            return SwiftReadFile(self.swiftfilesystem, fullpath)

        def err(failure):
            failure.trap(NotFound)
            return defer.fail(FileNotFoundError(fullpath))

        try:
            d = self.swiftfilesystem.checkFileExistance(fullpath)
            d.addCallback(cb)
            d.addErrback(err)
            return d
        except NotImplementedError:
            return defer.fail(IsADirectoryError(fullpath))

    def openForWriting(self, path):
        self.log_command('openForWriting', path)
        fullpath = self._fullpath(path)
        container, obj = obj_to_path(fullpath)
        if not container or not obj:
            log.msg('cannot upload to root')
            raise CmdNotImplementedForArgError(
                'Cannot upload files to root directory.')
        f = SwiftWriteFile(self.swiftfilesystem, fullpath)
        return defer.succeed(f)


class SwiftWriteFile:
    implements(IWriteFile)

    def __init__(self, swiftfilesystem, fullpath):
        self.swiftfilesystem = swiftfilesystem
        self.fullpath = fullpath
        self.finished = None

    def receive(self):
        d, writer = self.swiftfilesystem.startFileUpload(self.fullpath)
        self.finished = d
        return writer.started

    def close(self):
        return self.finished


class SwiftReadFile(Protocol):
    implements(IReadFile)

    def __init__(self, swiftfilesystem, fullpath):
        self.swiftfilesystem = swiftfilesystem
        self.fullpath = fullpath
        self.finished = defer.Deferred()

    def cb_send(self, result):
        return self.finished

    def send(self, consumer):
        self.consumer = consumer
        d = self.swiftfilesystem.startFileDownload(self.fullpath, self)
        d.addCallback(self.cb_send)
        return d

    def dataReceived(self, data):
        self.consumer.write(data)
        log.msg(metric='egress_bytes', count=len(data))

    def connectionLost(self, reason):
        from twisted.web._newclient import ResponseDone
        from twisted.web.http import PotentialDataLoss

        if reason.check(ResponseDone) or reason.check(PotentialDataLoss):
            self.finished.callback(None)
        else:
            self.finished.errback(reason)
        self.consumer.unregisterProducer()

    def makeConnection(self, transport):
        pass

    def connectionMade(self):
        pass



"""
This file contains the primary server code for the SFTP server.

See COPYING for license information.
"""
from zope import interface

from twisted.conch.interfaces import ISFTPServer, ISession
from twisted.cred import portal
from twisted.python import components, log
from twisted.internet import defer

from twisted.conch import avatar
from twisted.conch.ssh import session
from twisted.conch.ssh.filetransfer import FileTransferServer, SFTPError, \
    FX_FAILURE, FX_NO_SUCH_FILE
from twisted.conch.ssh.common import getNS
from twisted.conch.ssh.transport import SSHServerTransport

from swftp.swift import NotFound, Conflict
from swftp.sftp.swiftfile import SwiftFile
from swftp.sftp.swiftdirectory import SwiftDirectory
from swftp.swiftfilesystem import SwiftFileSystem, swift_stat, obj_to_path


class SwiftSession:
    " Barebones Session that closes when a client tries to open a shell "
    interface.implements(ISession)

    def __init__(self, avatar):
        self.avatar = avatar

    def openShell(self, proto):
        self.avatar.conn.transport.transport.loseConnection()

    def getPty(self, term, windowSize, modes):
        pass

    def closed(self):
        pass


class SwiftSFTPRealm:
    "Swift SFTP Realm"
    interface.implements(portal.IRealm)

    def requestAvatar(self, avatarId, mind, *interfaces):
        avatar = SwiftSFTPAvatar(avatarId)
        return interfaces[0], avatar, avatar.logout


class SwiftFileTransferServer(FileTransferServer):
    # Overridden to expose the session to the file object to do intellegent
    # throttling. Without this, memory bloat occurs.
    def _cbOpenFile(self, fileObj, requestId):
        fileObj.session = self.transport.session
        FileTransferServer._cbOpenFile(self, fileObj, requestId)

    # This is overridden because Flow was sending data that looks to be invalid
    def packet_REALPATH(self, data):
        requestId = data[:4]
        data = data[4:]
        path, data = getNS(data)
        # assert data == '', 'still have data in REALPATH: %s' % repr(data)
        d = defer.maybeDeferred(self.client.realPath, path)
        d.addCallback(self._cbReadLink, requestId)  # same return format
        d.addErrback(self._ebStatus, requestId, 'realpath failed')


class SwiftSSHServerTransport(SSHServerTransport):
    # Overridden to set the version string.
    version = 'SwFTP'
    ourVersionString = 'SSH-2.0-SwFTP'


class SwiftSFTPAvatar(avatar.ConchUser):
    "Swift SFTP Avatar"
    def __init__(self, swiftconn):
        avatar.ConchUser.__init__(self)
        self.swiftconn = swiftconn

        self.channelLookup.update({"session": session.SSHSession})
        self.subsystemLookup.update({"sftp": SwiftFileTransferServer})

        self.cwd = ''

    def logout(self):
        self.log_command('logout')
        self.swiftconn = None

    def log_command(self, command, *args):
        arg_list = ', '.join(str(arg) for arg in args)
        log.msg("COMMAND: %s(%s)" % (command, arg_list),
                system="SwFTP-SFTP, (%s)" % self.swiftconn.username,
                metric='command.%s' % command)


class SFTPServerForSwiftConchUser:
    "SFTP Server For a Swift User"
    interface.implements(ISFTPServer)

    def __init__(self, avatar):
        self.swiftconn = avatar.swiftconn
        self.swiftfilesystem = SwiftFileSystem(self.swiftconn)
        self.avatar = avatar
        self.conn = avatar.conn
        self.log_command('login')

    def log_command(self, *args, **kwargs):
        return self.avatar.log_command(*args, **kwargs)

    def gotVersion(self, otherVersion, extData):
        return {}

    def openFile(self, fullpath, flags, attrs):
        self.log_command('openFile', fullpath, flags, attrs)
        f = SwiftFile(self, fullpath, flags=flags, attrs=attrs)
        d = f.checkExistance()

        def errback(failure):
            failure.trap(NotFound)
            raise SFTPError(FX_FAILURE, "Container Doesn't Exist")

        d.addCallback(lambda r: f)
        d.addErrback(errback)
        return d

    def removeFile(self, fullpath):
        self.log_command('removeFile', fullpath)
        return self.swiftfilesystem.removeFile(fullpath)

    def renameFile(self, oldpath, newpath):
        self.log_command('renameFile', oldpath, newpath)
        d = self.swiftfilesystem.renameFile(oldpath, newpath)

        def errback(failure):
            failure.trap(NotFound, Conflict)
            if failure.check(NotFound):
                raise SFTPError(FX_NO_SUCH_FILE, 'No Such File')
            if failure.check(Conflict):
                raise NotImplementedError

        d.addErrback(errback)
        return d

    def makeDirectory(self, fullpath, attrs):
        self.log_command('makeDirectory', fullpath, attrs)

        def errback(failure):
            failure.trap(NotFound)
            raise SFTPError(FX_NO_SUCH_FILE, 'Directory Not Found')

        d = self.swiftfilesystem.makeDirectory(fullpath, attrs)
        d.addErrback(errback)
        return d

    def removeDirectory(self, fullpath):
        self.log_command('removeDirectory', fullpath)
        d = self.swiftfilesystem.removeDirectory(fullpath)

        def errback(failure):
            failure.trap(NotFound, Conflict)
            if failure.check(NotFound):
                return
            if failure.check(Conflict):
                raise SFTPError(FX_FAILURE, 'Directory Not Empty')

        d.addErrback(errback)
        return d

    def openDirectory(self, fullpath):
        self.log_command('openDirectory', fullpath)
        directory = SwiftDirectory(self.swiftfilesystem, fullpath)

        def cb(*result):
            return directory

        def errback(failure):
            failure.trap(NotFound)
            raise SFTPError(FX_FAILURE, 'Not Found')

        d = directory.get_full_listing()
        d.addCallback(cb)
        d.addErrback(errback)
        return d

    def getAttrs(self, fullpath, followLinks=False):
        self.log_command('getAttrs', fullpath)
        d = self.swiftfilesystem.getAttrs(fullpath)

        def cb(result):
            return self.format_attrs(result)

        def errback(failure):
            failure.trap(NotFound)
            raise SFTPError(FX_NO_SUCH_FILE, 'Not Found')

        d.addCallback(cb)
        d.addErrback(errback)

        return d

    def format_attrs(self, result):
        s = swift_stat(**result)
        return {
            "size": s.st_size,
            "uid": s.st_uid,
            "gid": s.st_gid,
            "permissions": s.st_mode,
            "atime": int(s.st_atime),
            "mtime": int(s.st_mtime)
        }

    def setAttrs(self, path, attrs):
        return

    def readLink(self, path):
        raise NotImplementedError

    def makeLink(self, linkPath, targetPath):
        raise NotImplementedError

    def realPath(self, path):
        container, obj = obj_to_path(path)
        real_path = '/'
        if container:
            real_path += container
        if obj:
            real_path += '/' + obj
        return real_path

    def extendedRequest(self, extName, extData):
        raise NotImplementedError

components.registerAdapter(
    SFTPServerForSwiftConchUser, SwiftSFTPAvatar, ISFTPServer)
components.registerAdapter(SwiftSession, SwiftSFTPAvatar, ISession)



"""
See COPYING for license information.
"""
from zope.interface import implements
from twisted.internet import defer
from twisted.cred import checkers, error, credentials

from swftp.swift import SwiftConnection, UnAuthenticated, UnAuthorized
from swftp.utils import USER_AGENT


class SwiftBasedAuthDB:
    """
        Swift-based authentication.

        Implements twisted.cred.ICredentialsChecker
    """
    implements(checkers.ICredentialsChecker)

    def __init__(self, auth_url=None, pool=None, verbose=False):
        self.auth_url = auth_url
        self.pool = pool
        self.verbose = verbose

    credentialInterfaces = (
        credentials.IUsernamePassword,
    )

    def _after_auth(self, result, connection):
        return connection

    def requestAvatarId(self, c):
        creds = credentials.IUsernamePassword(c, None)
        if creds is not None:
            conn = SwiftConnection(
                self.auth_url, creds.username, creds.password,
                pool=self.pool,
                verbose=self.verbose)
            conn.user_agent = USER_AGENT
            d = conn.authenticate()
            d.addCallback(self._after_auth, conn)
            d.addErrback(failed_auth)
            return d
        return defer.fail(error.UnauthorizedLogin())


def failed_auth(failure):
    failure.trap(UnAuthenticated, UnAuthorized)
    return defer.fail(error.UnauthorizedLogin())



"""
See COPYING for license information.
"""
from twisted.conch import ls

from swftp.utils import OrderedDict
from swftp.swiftfilesystem import swift_stat


class SwiftDirectory(object):
    "Swift Directory is an iterator that returns a listing of the directory."
    def __init__(self, swiftfilesystem, fullpath):
        self.swiftfilesystem = swiftfilesystem
        self.fullpath = fullpath
        # A lot of files require . and .. to be within the directory listing
        self.files = OrderedDict(
            [
                ('.', {}),
                ('..', {}),
            ])
        self.done = False

    def get_full_listing(self):
        "Populate the directory listing."
        def cb(results):
            for k, v in results.iteritems():
                self.files[k] = v

        d = self.swiftfilesystem.get_full_listing(self.fullpath)
        d.addCallback(cb)
        return d

    def __iter__(self):
        return self

    def next(self):
        try:
            name, f = self.files.popitem(last=False)
            lstat = swift_stat(**f)
            longname = ls.lsLine(name, lstat)
            return (name, longname, {
                "size": lstat.st_size,
                "uid": lstat.st_uid,
                "gid": lstat.st_gid,
                "permissions": lstat.st_mode,
                "atime": int(lstat.st_atime),
                "mtime": int(lstat.st_mtime)
            })
        except KeyError:
            raise StopIteration

    def close(self):
        self.files = []
        self.offset = 0



"""
See COPYING for license information.
"""
from syslog import (
    LOG_USER, LOG_DAEMON, LOG_SYSLOG, LOG_LOCAL0, LOG_LOCAL1, LOG_LOCAL2,
    LOG_LOCAL3, LOG_LOCAL4, LOG_LOCAL5, LOG_LOCAL6, LOG_LOCAL7)

from twisted.python import syslog


class SysLogObserver:
    facility = LOG_USER

    def __init__(self):
        self.obs = syslog.SyslogObserver('swftp', facility=self.facility)

    def __call__(self, event_dict):
        self.obs.emit(event_dict)


class LOG_USER(SysLogObserver):
    facility = LOG_USER


class LOG_DAEMON(SysLogObserver):
    facility = LOG_DAEMON


class LOG_SYSLOG(SysLogObserver):
    facility = LOG_SYSLOG


class LOG_LOCAL0(SysLogObserver):
    facility = LOG_LOCAL0


class LOG_LOCAL1(SysLogObserver):
    facility = LOG_LOCAL1


class LOG_LOCAL2(SysLogObserver):
    facility = LOG_LOCAL2


class LOG_LOCAL3(SysLogObserver):
    facility = LOG_LOCAL3


class LOG_LOCAL4(SysLogObserver):
    facility = LOG_LOCAL4


class LOG_LOCAL5(SysLogObserver):
    facility = LOG_LOCAL5


class LOG_LOCAL6(SysLogObserver):
    facility = LOG_LOCAL6


class LOG_LOCAL7(SysLogObserver):
    facility = LOG_LOCAL7



"""
See COPYING for license information.
"""
import time
import signal

import twisted.internet.tcp
from twisted.python import log
from twisted.internet import reactor
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # NOQA

VERSION = '1.0.5'
USER_AGENT = 'SwFTP v%s' % VERSION

DATE_FORMATS = [
    "%a, %d %b %Y %H:%M:%S %Z",
    "%a, %d %b %Y %H:%M:%S.%f %Z",
    "%Y-%m-%dT%H:%M:%S",
    "%Y-%m-%dT%H:%M:%S.%f",
    "%Y-%m-%d %H:%M:%S",
    "%Y-%m-%d %H:%M:%S.%f",
    "%Y-%m-%d"
]


def try_datetime_parse(datetime_str):
    """
    Tries to parse the datetime and return the UNIX epoch version of the time.

    returns timestamp(float) or None
    """
    mtime = None
    if datetime_str:
        for format in DATE_FORMATS:
            try:
                mtime_tuple = time.strptime(datetime_str, format)
                mtime = time.mktime(tuple(mtime_tuple))
            except ValueError:
                pass
            else:
                break
    return mtime


def print_runtime_info(sig, frame):
    if sig in [signal.SIGUSR1, signal.SIGUSR2]:
        delayed = reactor.getDelayedCalls()
        readers = reactor.getReaders()
        writers = reactor.getWriters()
        clients = []
        http_conn_num = 0
        for reader in readers:
            if isinstance(reader, twisted.internet.tcp.Server):
                clients.append(reader.getPeer())
            if isinstance(reader, twisted.internet.tcp.Client):
                http_conn_num += 1
        log.msg("[Clients: %(client_num)s] [HTTP Conns: %(http_conn_num)s] "
                "[Readers: %(reader_num)s] [Writers: %(writer_num)s] "
                "[DelayedCalls: %(delayed_num)s]" % {
                    "client_num": len(clients),
                    "http_conn_num": http_conn_num,
                    "reader_num": len(readers),
                    "writer_num": len(writers),
                    "delayed_num": len(delayed),
                })
        log.msg("[Connected Clients]: %s" % clients)
        if sig == signal.SIGUSR2:
            for d in delayed:
                log.msg("SIGUSR2[delayed]: %s" % d)

            for r in readers:
                log.msg("SIGUSR2[reader]: %s" % r)

            for w in writers:
                log.msg("SIGUSR2[writer]: %s" % w)



"""
See COPYING for license information.
"""
from twisted.conch.ssh.connection import SSHConnection, \
    MSG_CHANNEL_WINDOW_ADJUST
import struct


# SSHConnection is overridden to reduce verbosity.
class SwiftConnection(SSHConnection):
    def adjustWindow(self, channel, bytesToAdd):
        if channel.localClosed:
            return  # we're already closed
        self.transport.sendPacket(MSG_CHANNEL_WINDOW_ADJUST, struct.pack('>2L',
                                  self.channelsToRemoteChannel[channel],
                                  bytesToAdd))
        channel.localWindowLeft += bytesToAdd



"""
SwFTP is an FTP and SFTP interface for Openstack Swift

See COPYING for license information.
"""
from swftp.utils import VERSION
__title__ = 'swftp'
__version__ = VERSION
__author__ = 'SoftLayer Technologies, Inc.'
__license__ = 'MIT'
__copyright__ = 'Copyright 2013 SoftLayer Technologies, Inc.'




class SwiftSSHServerTransport(SSHServerTransport):
    # Overridden to set the version string.
    version = 'SwFTP'
    ourVersionString = 'SSH-2.0-SwFTP'

    def __init__(self):
        self._last_seq_num = 0
        self._idle_failed_count = 0
        self._idle_check = task.LoopingCall(self.__check_idle)
        self._idle_check.start(900)

    def __check_idle(self):
        if self.incomingPacketSequence == self._last_seq_num:
            self._idle_failed_count += 1
            if self._idle_failed_count >= 2:
                self.sendDisconnect(DISCONNECT_CONNECTION_LOST, "idle timeout")
                self._idle_check.stop()
        self._last_seq_num = self.incomingPacketSequence

    def dataReceived(self, data):
        if data.strip() in ['quit', 'exit']:
            self.sendDisconnect(DISCONNECT_CONNECTION_LOST, "exit")
        return SSHServerTransport.dataReceived(self, data)


#!/usr/bin/python
from setuptools import setup
import sys

name = 'swftp'

short_description = 'SwFTP is an FTP and SFTP interface for Openstack Swift'
long_description = short_description
try:
    long_description = open('README.md').read()
except:
    pass

requires = ['twisted >= 12.1', 'pyopenssl', 'pycrypto', 'pyasn1', 'psutil']

if sys.version_info < (2, 7):
    requires.append('ordereddict')

setup(
    name=name,
    version='1.0.5',
    author='Kevin McDonald',
    author_email='kmcdonald@softlayer.com',
    license='MIT',
    url='https://github.com/softlayer/swftp',
    description=short_description,
    long_description=long_description,
    packages=['swftp', 'swftp/sftp', 'swftp/ftp', 'twisted.plugins'],
    install_requires=requires,
    entry_points={
        'console_scripts': ['swftp-ftp = swftp.ftp.service:run',
                            'swftp-sftp = swftp.sftp.service:run'],
    },
    package_data={
        'twisted.plugins': ['twisted/plugins/swftp_ftp.py',
                            'twisted/plugins/swftp_sftp.py']},
    data_files=[
        ('', ['README.md']),
        ('/etc/swftp', ['etc/swftp/swftp.conf.sample']),
    ],
    zip_safe=False,
    classifiers=[
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Environment :: No Input/Output (Daemon)',
        'Framework :: Twisted',
        'License :: OSI Approved :: MIT License',
        'Topic :: Internet :: File Transfer Protocol (FTP)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
)

# Make Twisted regenerate the dropin.cache, if possible.  This is necessary
# because in a site-wide install, dropin.cache cannot be rewritten by
# normal users.
try:
    from twisted.plugin import IPlugin, getPlugins
except ImportError:
    pass
else:
    list(getPlugins(IPlugin))



"""
This file defines what is required for swftp-sftp to work with twistd.

See COPYING for license information.
"""
from swftp.utils import VERSION

from twisted.application import internet, service
from twisted.python import usage, log
from twisted.internet import reactor

import ConfigParser
import signal
import os
import sys
import time


def run():
    options = Options()
    try:
        options.parseOptions(sys.argv[1:])
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)
    log.startLogging(sys.stdout)
    s = makeService(options)
    s.startService()
    reactor.run()


def get_config(config_path, overrides):
    defaults = {
        'auth_url': 'http://127.0.0.1:8080/auth/v1.0',
        'host': '0.0.0.0',
        'port': '5022',
        'priv_key': '/etc/swftp/id_rsa',
        'pub_key': '/etc/swftp/id_rsa.pub',
        'num_persistent_connections': '4',
        'connection_timeout': '240',
        'verbose': 'false',
        'log_statsd_host': '',
        'log_statsd_port': '8125',
        'log_statsd_sample_rate': '10.0',
        'log_statsd_metric_prefix': 'swftp.sftp',
    }
    c = ConfigParser.ConfigParser(defaults)
    c.add_section('sftp')
    if config_path:
        log.msg('Reading configuration from path: %s' % config_path)
        c.read(config_path)
    else:
        config_paths = [
            '/etc/swftp/swftp.conf',
            os.path.expanduser('~/.swftp.cfg')
        ]
        log.msg('Reading configuration from paths: %s' % config_paths)
        c.read(config_paths)
    for k, v in overrides.iteritems():
        if v:
            c.set('sftp', k, str(v))
    return c


class Options(usage.Options):
    "Defines Command-line options for the swftp-sftp service"
    optFlags = [
        ["verbose", "v", "Make the server more talkative"]
    ]
    optParameters = [
        ["config_file", "c", None, "Location of the swftp config file."],
        ["auth_url", "a", None,
            "Auth Url to use. Defaults to the config file value if it exists."
            "[default: http://127.0.0.1:8080/auth/v1.0]"],
        ["port", "p", None, "Port to bind to."],
        ["host", "h", None, "IP to bind to."],
        ["priv_key", "priv-key", None, "Private Key Location."],
        ["pub_key", "pub-key", None, "Public Key Location."],
    ]


def makeService(options):
    """
    Makes a new swftp-sftp service. The only option is the config file
    location. The config file has the following options:
     - host
     - port
     - auth_url
     - num_persistent_connections
     - connection_timeout
     - pub_key
     - priv_key
    """
    from twisted.conch.ssh.factory import SSHFactory
    from twisted.conch.ssh.keys import Key
    from twisted.web.client import HTTPConnectionPool
    from twisted.cred.portal import Portal

    from swftp.sftp.server import SwiftSFTPRealm, SwiftSSHServerTransport
    from swftp.sftp.connection import SwiftConnection
    from swftp.auth import SwiftBasedAuthDB
    from swftp.utils import print_runtime_info

    c = get_config(options['config_file'], options)
    pool = HTTPConnectionPool(reactor, persistent=True)
    pool.maxPersistentPerHost = c.getint('sftp', 'num_persistent_connections')
    pool.cachedConnectionTimeout = c.getint('sftp', 'connection_timeout')

    sftp_service = service.MultiService()

    # ensure timezone is GMT
    os.environ['TZ'] = 'GMT'
    time.tzset()

    log.msg('Starting SwFTP-sftp %s' % VERSION)

    # Add statsd service
    if c.get('sftp', 'log_statsd_host'):
        try:
            from swftp.statsd import makeService as makeStatsdService
            makeStatsdService(
                c.get('sftp', 'log_statsd_host'),
                c.getint('sftp', 'log_statsd_port'),
                sample_rate=c.getfloat('sftp', 'log_statsd_sample_rate'),
                prefix=c.get('sftp', 'log_statsd_metric_prefix')
            ).setServiceParent(sftp_service)
        except ImportError:
            log.err('Missing Statsd Module. Requires "txstatsd"')

    authdb = SwiftBasedAuthDB(auth_url=c.get('sftp', 'auth_url'),
                              verbose=c.getboolean('sftp', 'verbose'))

    sftpportal = Portal(SwiftSFTPRealm())
    sftpportal.registerChecker(authdb)

    sshfactory = SSHFactory()
    sshfactory.protocol = SwiftSSHServerTransport
    sshfactory.noisy = False
    sshfactory.portal = sftpportal
    sshfactory.services['ssh-connection'] = SwiftConnection

    pub_key_string = file(c.get('sftp', 'pub_key')).read()
    priv_key_string = file(c.get('sftp', 'priv_key')).read()
    sshfactory.publicKeys = {
        'ssh-rsa': Key.fromString(data=pub_key_string)}
    sshfactory.privateKeys = {
        'ssh-rsa': Key.fromString(data=priv_key_string)}

    signal.signal(signal.SIGUSR1, print_runtime_info)
    signal.signal(signal.SIGUSR2, print_runtime_info)

    internet.TCPServer(
        c.getint('sftp', 'port'),
        sshfactory,
        interface=c.get('sftp', 'host')).setServiceParent(sftp_service)

    return sftp_service
